
/*  sff.c:  simple force field: implement routines to read a "prmtop"
 *      file from AMBER, and calculate the energy.  Implements only
 *      some of AMBER's functionality: bonds, angles, dihedrals, and
 *      nonbonded interactions with a distance-dependent dielectric.
 *      Does not (yet) include support for period systems.
 *
 *      Main interface is through routines "mme" and "mme_init".
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "nab.h"
#include "memutil.h"
#include "debug.h"

						   /*  general force field variables   */

#ifdef flex
static char *mmoinputptr;
static int mmoinputlim;
#endif

static REAL_T t1, t2, tnonb, tpair, tbond, tangl, tphi, tcons;
							/* timer variables  */

static char *gopts;				/*  points to the mm_options string  */
static REAL_T cut = 8.0;		/* non-bonded cutoff */
static int ntpr = 10;			/*  print frequency  */
static int nsnb = 25;			/*  non-bonded update frequency  */
static REAL_T scnb = 2.0;		/*  scale factor for 1-4 nonbonds  */
static REAL_T scee = 1.2;		/*  scale factor for 1-4 electrostatics  */

static REAL_T rgbmax = 999.0;   /*  maximum dist. for gb radius calculation */
static REAL_T gbalpha, gbbeta, gbgamma;

static int mme_init_first = 1;
static PARMSTRUCT_T *prm;

static int *frozen;
static int nfrozen = 0;

static int *constrained;
static int nconstrained = 0;
static REAL_T *x0;				/*  will hold coordinates to be constrained  */
static REAL_T wcons = 0.0;		/*  weight of constraints    */

static int *npairs;             /* array over atoms of the number of pairs */
static int *pairlist = NULL;    /* array over atoms of the list of pairs */
static int nb_pairs = -1;       /* the total number of pairs */

static int gb = 0;				/* non-zero if generalized Born is to be run */
static int gb_debug = 0;		/* non-zero if extra printout for gen. Born  */
static int e_debug = 0;			/* non-zero if extra printout for energies   */

/*LCPO stuff */
static REAL_T *P0, *P1, *P2, *P3, *P4;	/* augmented VdW radii, parameters (see JCC 20,2,217 ff.) */
static int *ineighbor;			/* neighborlist-array */
static int gbsa = 0;
static REAL_T surften = 0.005;	/* 'surface tension' for GB/SA */
static REAL_T esurf = 0.0;		/* nonpolar solvation */

static REAL_T epsext = 78.5;	/* exterior dielectric constant for GB */
static REAL_T kappa = 0.0;		/* Debye-Huckel screeing parameter for GB */

static int maxnb;
static int dield = 1;			/* dielectric function to be used */
static REAL_T k4d = 0.0;
static int dim;					/* dimension: 3 or 4 */


							/*  MD variables:  */

static REAL_T t = 0.;			/* initial time */
static REAL_T dt = 0.001;		/* time step, ps. */
static REAL_T tautp = 999999.0;	/* temp. coupling parm., ps */
static REAL_T gamma_ln = 0.0;   /* Langevin collison parameter, in ps^-1  */
static REAL_T temp0 = 300.;		/* target temperature, K */
static REAL_T boltz2 = 9.93595e-4;
static REAL_T vlimit = 10.0;	/* maximum velocity component */
static REAL_T genmass = 10.;	/* general masses, for now all the same */
static int ntpr_md = 10;		/* print frequency for KE,PE,temp */
static int ntwx = 0;			/* trajectory snapshot frequency  */
static FILE *binposfp;			/* file pointer for trajectories  */
static int zerov = 0;			/* if true, use zero initial velocities */
static REAL_T tempi = 0.0;		/* initial temperature */
static int idum = -1;			/* random number seed */

static int irattle = 0;			/* turn on/off bond constraints in md */

REAL_T second();
REAL_T gauss();

/*   The energy functions themselves:   */
#include "eff.c"
#include "sff4.c"

int	mme_init( MOLECULE_T *, char *, char *, REAL_T *, FILE * );
REAL_T	mme( REAL_T *, REAL_T *, int * );
REAL_T	mme_rattle( REAL_T *, REAL_T *f, int * );
int	get_masses( REAL_T * );
int	mm_options( char * );
int	md( int,  int, REAL_T *, REAL_T *, REAL_T *, REAL_T (*)() );
int	mme_timer( void );

int	mme_init( MOLECULE_T *m, char *aexp, char *aexp2, REAL_T *x0i, FILE *bfpi )
/*
MOLECULE_T *m;
char *aexp, *aexp2;
REAL_T *x0i;
FILE *bfpi;
*/
{

	int i, n2o2, j;

	int at1, at2;
	char atypi[3];
	int *numb;

	dim = 3;
	prm = m->m_prm;
	x0 = x0i;
	binposfp = bfpi;

	if (mme_init_first) {

		if (gbsa == 1) {
			numb = ivector(0, prm->Natom);
			for (i = 0; i < prm->Natom; i++)
				numb[i] = 0;

			for (i = 0; i < prm->Nbona; i++) {
				at1 = prm->BondAt1[i] / 3;
				at2 = prm->BondAt2[i] / 3;
				numb[at1] += 1;
				numb[at2] += 1;
			}

			P0 = vector(0, prm->Natom);
			P1 = vector(0, prm->Natom);
			P2 = vector(0, prm->Natom);
			P3 = vector(0, prm->Natom);
			P4 = vector(0, prm->Natom);
			ineighbor = ivector(0, 30 * prm->Natom);
			/* 30 seems sufficient. this should, however, be changed */
			/* if it proves unsafe */
			/* mabye to size 80000, which is well beyond what's needed */
			/* for a whole tRNA */



			for (i = 0; i < prm->Natom; i++) {
				atypi[0] = prm->AtomSym[i * 4];
				atypi[1] = prm->AtomSym[i * 4 + 1];
				atypi[2] = 0;


				if (strncmp(atypi, "CT", 2) == 0) {
					if (numb[i] == 1) {

						P0[i] = 1.70 + 1.40;
						P1[i] = 0.7887;
						P2[i] = -0.28063;
						P3[i] = -0.0012968;
						P4[i] = 0.00039328;
					} else if (numb[i] == 2) {
						P0[i] = 1.70 + 1.4;
						P1[i] = 0.56482;
						P2[i] = -0.19608;
						P3[i] = -0.0010219;
						P4[i] = 0.0002658;
					}

					else if (numb[i] == 3) {
						P0[i] = 1.70 + 1.4;
						P1[i] = 0.23348;
						P2[i] = -0.072627;
						P3[i] = -0.00020079;
						P4[i] = 0.00007967;
					}

					else if (numb[i] == 4) {
						P0[i] = 1.70 + 1.4;
						P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
					}
				} else if (strncmp(atypi, "C", 1) == 0) {
					if (numb[i] == 2) {
						P0[i] = 1.70 + 1.4;
						P1[i] = 0.51245;
						P2[i] = -0.15966;
						P3[i] = -0.00019781;
						P4[i] = 0.00016392;
					} else if (numb[i] == 3) {
						P0[i] = 1.70 + 1.4;
						P1[i] = 0.070344;
						P2[i] = -0.019015;
						P3[i] = -0.000022009;
						P4[i] = 0.000016875;
					}
				} else if (strncmp(atypi, "O ", 2) == 0) {
					P0[i] = 1.6 + 1.4;
					P1[i] = 0.68563;
					P2[i] = -0.1868;
					P3[i] = -0.00135573;
					P4[i] = 0.00023743;
				} else if (strncmp(atypi, "O2", 2) == 0) {
					P0[i] = 1.6 + 1.4;
					P1[i] = 0.88857;
					P2[i] = -0.33421;
					P3[i] = -0.0018683;
					P4[i] = 0.00049372;
				} else if (strncmp(atypi, "O", 1) == 0) {
					if (numb[i] == 1) {
						P0[i] = 1.6 + 1.4;
						P1[i] = 0.77914;
						P2[i] = -0.25262;
						P3[i] = -0.0016056;
						P4[i] = 0.00035071;
					} else if (numb[i] == 2) {
						P0[i] = 1.6 + 1.4;
						P1[i] = 0.49392;
						P2[i] = -0.16038;
						P3[i] = -0.00015512;
						P4[i] = 0.00016453;
					}
				} else if (strncmp(atypi, "N3", 2) == 0) {
					if (numb[i] == 1) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.078602;
						P2[i] = -0.29198;
						P3[i] = -0.0006537;
						P4[i] = 0.00036247;
					} else if (numb[i] == 2) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.22599;
						P2[i] = -0.036648;
						P3[i] = -0.0012297;
						P4[i] = 0.000080038;
					} else if (numb[i] == 3) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.051481;
						P2[i] = -0.012603;
						P3[i] = -0.00032006;
						P4[i] = 0.000024774;
					}
				} else if (strncmp(atypi, "N", 1) == 0) {
					if (numb[i] == 1) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.73511;
						P2[i] = -0.22116;
						P3[i] = -0.00089148;
						P4[i] = 0.0002523;
					} else if (numb[i] == 2) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.41102;
						P2[i] = -0.12254;
						P3[i] = -0.000075448;
						P4[i] = 0.00011804;
					} else if (numb[i] == 3) {
						P0[i] = 1.65 + 1.4;
						P1[i] = 0.062577;
						P2[i] = -0.017874;
						P3[i] = -0.00008312;
						P4[i] = 0.000019849;
					}
				} else if (strncmp(atypi, "SH", 2) == 0) {
					P0[i] = 1.9 + 1.4;
					P1[i] = 0.7722;
					P2[i] = -0.26393;
					P3[i] = 0.0010629;
					P4[i] = 0.0002179;
				} else if (strncmp(atypi, "S", 1) == 0) {
					P0[i] = 1.9 + 1.4;
					P1[i] = 0.54581;
					P2[i] = -0.19477;
					P3[i] = -0.0012873;
					P4[i] = 0.00029247;
				} else if (strncmp(atypi, "P", 1) == 0) {
					if (numb[i] == 3) {
						P0[i] = 1.9 + 1.4;
						P1[i] = 0.3865;
						P2[i] = -0.18249;
						P3[i] = -0.0036598;
						P4[i] = 0.0004264;
					} else if (numb[i] == 4) {
						P0[i] = 1.9 + 1.4;
						P1[i] = 0.03873;
						P2[i] = -0.0089339;
						P3[i] = 0.0000083582;
						P4[i] = 0.0000030381;
					}
				} else if (strncmp(atypi, "H", 1) == 0) {
					P0[i] = 1.40;
					P1[i] = P2[i] = P3[i] = P4[i] = 0.0;
				} else {
					fprintf(stderr, "Unknown Atom %d found, bailing out\n",
							i);
					exit(1);
				}

			}
			free_ivector(numb, 0, prm->Natom);

		}
		/*GBSA initialization ends here */

		frozen = ivector(0, prm->Natom);
		constrained = ivector(0, prm->Natom);
		npairs = ivector(0, prm->Natom);

		if (binposfp != NULL)
			writebinposhdr(binposfp);
		mme_init_first = 0;
	}

	if( !gb ){
		/* estimate for the number of non-bonded pairs */
	    float n2o2, natom_float, maxnb_float;
		if (pairlist != NULL)
			free_ivector(pairlist, 0, maxnb);

		natom_float = prm->Natom;
		maxnb_float = natom_float * cut * cut * cut / 1.25;
		n2o2 = natom_float * (natom_float - 1.) / 2.;
		maxnb_float = maxnb_float < n2o2 ? maxnb_float : n2o2;

		if (maxnb_float > 2.147e9){
		    printf("Unreasonably large number of pairs: %10.4e\n", maxnb_float);
			exit(1);
		} else {
			maxnb = maxnb_float;
			printf("allocating space for %u non-bonded pairs\n", maxnb);
			pairlist = ivector(0, maxnb);
		}
	}

	nfrozen = set_belly_mask(m, aexp, frozen);
	if (nfrozen)
		printf("froze %d atoms using expression %s for moving atoms\n",
			   nfrozen, aexp);

	nconstrained = set_cons_mask(m, aexp2, constrained);
	if (nconstrained)
		if (aexp2 == NULL)
			printf("constrained all %d atoms\n", nconstrained);
		else
			printf("constrained %d atoms using expression %s\n",
				   nconstrained, aexp2);

	/*  set up GB/OBC parameters:  */
	if( gb==2 ){
		gbalpha = 0.8; gbbeta = 0.0; gbgamma = 2.909125;
	} else if ( gb==5 ){
		gbalpha = 1.0; gbbeta = 0.8; gbgamma = 4.85;
	}

	return (0);
}

REAL_T	mme( REAL_T *x,  REAL_T *f, int *iter )
/*
REAL_T *x, *f;
int *iter;
*/
{
	REAL_T ebh, eba, eth, eta, eph, epa, enb, eel, enb14, eel14, ecn, frms;
	REAL_T ene[20];
	int i, k;

	t1 = second();
	if (*iter == 0) {
		printf
			("    iter    Total    bad         vdW     elect.     cons.   genBorn      Esurf\n");
		tnonb = tpair = tbond = tangl = tphi = tcons = 0.0;
	}
#if 0
	for( i=0; i<5; i++ ) printf( "%d %d %d %d %d\n",  prm->DihHAt1[i],
		prm->DihHAt2[i], prm->DihHAt3[i],
		prm->DihHAt4[i], prm->DihHNum[i] );
#endif
	if (!gb && (nb_pairs < 0 || (*iter % nsnb == 0 && *iter != 0))) {
		int firstlist = nb_pairs < 0;
		nb_pairs = nblist(x, npairs, pairlist);
		t2 = second();
		tpair += t2 - t1;
		t1 = t2;
	}

	for (i = 0; i < 3 * prm->Natom; i++)
		f[i] = 0.0;

	ebh = ebond(prm->Nbonh, prm->BondHAt1, prm->BondHAt2,
				prm->BondHNum, prm->Rk, prm->Req, x, f);
	eba = ebond(prm->Mbona, prm->BondAt1, prm->BondAt2,
				prm->BondNum, prm->Rk, prm->Req, x, f);
	ene[3] = ebh + eba;
	t2 = second();
	tbond += t2 - t1;
	t1 = t2;

	eth = eangl(prm->Ntheth, prm->AngleHAt1, prm->AngleHAt2,
				prm->AngleHAt3, prm->AngleHNum, prm->Tk, prm->Teq, x, f);
	eta = eangl(prm->Ntheta, prm->AngleAt1, prm->AngleAt2,
				prm->AngleAt3, prm->AngleNum, prm->Tk, prm->Teq, x, f);
	ene[4] = eth + eta;
	t2 = second();
	tangl += t2 - t1;
	t1 = t2;

	eph = ephi(prm->Nphih, prm->DihHAt1, prm->DihHAt2,
			   prm->DihHAt3, prm->DihHAt4, prm->DihHNum,
			   prm->Pk, prm->Pn, prm->Phase, x, f);
	epa = ephi(prm->Mphia, prm->DihAt1, prm->DihAt2,
			   prm->DihAt3, prm->DihAt4, prm->DihNum,
			   prm->Pk, prm->Pn, prm->Phase, x, f);
	ene[5] = eph + epa;

	ene[6] = 0.0;				/*  hbond term not in Amber-94 force field */
	nbond(prm->N14pairs, prm->N14pairlist, x, f, &enb14, &eel14,
		  scnb, scee);
	if (e_debug) {
		EXPR("%9.3f", enb14);
		EXPR("%9.3f", eel14);
	}
	ene[7] = enb14;
	ene[8] = eel14;
	t2 = second();
	tphi += t2 - t1;
	t1 = t2;

	if (nconstrained) {
		ecn = econs(x, f);
		t2 = second();
		tcons += t2 - t1;
		t1 = t2;
	} else
		ecn = 0.0;
	ene[9] = ecn;

	if (gb) {
		ene[10] = egb(&prm->Natom, x, f, prm->Fs, prm->Rborn, prm->Charges,
					  &kappa, &epsext, &enb, &eel);
		ene[1] = enb;
		ene[2] = eel;
	} else {
		nbond(npairs, pairlist, x, f, &enb, &eel, 1.0, 1.0);
		ene[1] = enb;
		ene[2] = eel;
		ene[10] = 0.0;
	}
	t2 = second();
	tnonb += t2 - t1;
	t1 = t2;

	ene[0] = 0.0;
	for (k = 1; k <= 10; k++)
		ene[0] += ene[k];

	for (k = 0; k < prm->Natom; k++) {	/* zero out frozen forces */
		if (frozen[k]) {
			f[3 * k + 0] = 0.0;
			f[3 * k + 1] = 0.0;
			f[3 * k + 2] = 0.0;
		}
	}

#ifdef PRINT_DERIV
	k = 0;
	for (i = 0; i < 105; i++) {
		k++;
		printf("%10.5f", f[i]);
		if (k % 8 == 0)
			printf("\n");
	}
	printf("\n");
#endif

#if 0
	frms = 0.0;
	for (i = 0; i < 3 * prm->Natom; i++)
		frms += f[i] * f[i];
	frms = sqrt(frms / (3 * prm->Natom));
#endif

	if (*iter == 0 || *iter % ntpr == 0) {
		printf("ff:%4d %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
			   *iter, ene[0] + esurf, ene[3] + ene[4] + ene[5],
			   ene[1] + ene[7], ene[2] + ene[8], ene[9], ene[10], esurf);
		fflush(stdout);
	}

	return (ene[0] + esurf);
}

REAL_T	mme_rattle( REAL_T *x, REAL_T *f, int *iter )
/*
REAL_T *x, *f;
int *iter;
*/
{
	REAL_T ene;


	/*
	 * Version of mme() with rattle for bond length constraints, for
	 * use in minimization. This is quite loosely based on  Duan, Kumar, 
	 * Rosenberg, and Kollman, J. Computat. Chem. 16, 1351-1356 (1995).
	 */

	int dumx, nratH, nrat, Trat, at1, at2, atyp;
	int nat3, i, done, niter, maxiter, imax;
	static REAL_T *minv, *dumv, *deltag;
	REAL_T dtx;
	REAL_T maxf, tmpf;
	REAL_T inP, krat, dist2, xr, yr, zr, delta;
	REAL_T bn, bnx, bny, bnz, frms;

	nat3 = 3 * prm->Natom;
	dtx = 0.001 * 20.455;
	maxiter = 1000;

	niter = 0;
	done = 0;

	/*
	 * inverse masses and velocities are required by the
	 * RATTLE routine that's called below, but are dummy variables
	 * since here only positions are affected
	 */

	if (!minv) {
		minv = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		dumv = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		deltag = (REAL_T *) calloc(nat3, sizeof(REAL_T));
		for (i = 0; i < nat3; i++) {
			minv[i] = 2.0;
			dumv[i] = 0.0;
		}
	}

	for (i = 0; i < nat3; i++)  deltag[i] = 0.0;
	maxf = -1.0;

	/* call the 'position part' of RATTLE to repair bond lengths */
	dumx = rattle(dtx, x, x, dumv, minv);

	/* get energy and gradient for the RATTLEd structure */
	ene = mme(x, f, iter);

	/* repair the gradient */

	nratH = prm->Nbonh;
	nrat = prm->Mbona;
	Trat = nratH + nrat;

	while ((done == 0) && (niter < 50)) {
		niter++;
		done = 1;
		for (i = 0; i < Trat; i++) {

			if (i < nratH) {
				at1 = prm->BondHAt1[i];
				at2 = prm->BondHAt2[i];
				atyp = prm->BondHNum[i] - 1;
				krat = prm->Req[atyp];
			} else {
				at1 = prm->BondAt1[i - nratH];
				at2 = prm->BondAt2[i - nratH];
				atyp = prm->BondNum[i - nratH] - 1;
				krat = prm->Req[atyp];
			}

			xr = x[at2] - x[at1];
			yr = x[at2 + 1] - x[at1 + 1];
			zr = x[at2 + 2] - x[at1 + 2];

			dist2 = xr * xr + yr * yr + zr * zr;
			/* 'inP' comes from the german expression for dot product */
			/* since thats exactly what's calculated here */
			inP =
				xr * (f[at1] - f[at2]) + yr * (f[at1 + 1] - f[at2 + 1]) +
				zr * (f[at1 + 2] - f[at2 + 2]);

#define EPS 0.0000001
			if (inP > EPS) done = 0;
			f[at1] -= (xr / (2.0 * dist2)) * inP;
			f[at1 + 1] -= (yr / (2.0 * dist2)) * inP;
			f[at1 + 2] -= (zr / (2.0 * dist2)) * inP;
			f[at2] += (xr / (2.0 * dist2)) * inP;
			f[at2 + 1] += (yr / (2.0 * dist2)) * inP;
			f[at2 + 2] += (zr / (2.0 * dist2)) * inP;
		}
	}

#ifdef DEBUG
	frms = 0.0;
	for (i = 0; i < nat3 / 3; i++) {
		tmpf =
			f[3 * i] * f[3 * i] + f[3 * i + 1] * f[3 * i + 1] + f[3 * i +
																  2] *
			f[3 * i + 2];
		frms += tmpf;

		if (sqrt(tmpf) > maxf) {
			maxf = sqrt(tmpf);
			imax = i;

		}
	}
	/* the last for loop is pretty inefficient ;-) */
	/* but the output is informative nonetheless */

	frms = sqrt(frms / (REAL_T) nat3);
	printf("Ene %18.14f frms %18.14f\n", ene, frms);
	printf("Max grad %f for atom %d\n", maxf, imax + 1);
	/* output accuracy is a little exaggerated */
	/* we want to see all those digits */
#endif

	return ene;
}

int	get_masses( REAL_T *minv )
									/* dim = 3 or 4, for 3D or 4D problem */
/*
 REAL_T *minv;
*/
{

	int i, k;
	REAL_T am;

	for (k = 0, i = 0; i < prm->Natom; i++) {
		am = 1. / prm->Masses[i];
		minv[k + 0] = am;
		minv[k + 1] = am;
		minv[k + 2] = am;
		if (dim == 4)
			minv[k + 3] = am;
		k += dim;
	}
	return (0);
}

int	mm_options( char *opts )
/*
char *opts;
*/
{
	gopts = opts;
#ifdef flex
	mmoinputlim = strlen(opts);
	mmoinputptr = gopts;
#endif
	mmolex();
	return (0);
}

int	md( int n,  int maxstep, REAL_T *x, REAL_T *f, REAL_T *v, REAL_T ( *mme )() )
/*
 REAL_T(*mme) ();
int n, maxstep;
REAL_T *x, *f, *v;
*/
{

	REAL_T dtx, dt5, rndf, dttp, ekin0, ekin, epot, etot, tscal, temp, sd;
	REAL_T gammai, c_implic, c_explic, sdfac;
	int nstep, i;
	REAL_T zero, invmass;
	static REAL_T *xold, *sqrmass, *accel, *minv;

	/*  space for holding inverse masses  */
	if( !minv ){
		minv = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( minv == NULL ){
			fprintf( stderr, "unable to allocate space for minv\n" );
			exit(1);
		}
	}

	/*  space for holding extra set of coords  */
	if( irattle && !xold ){
		xold = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( xold == NULL ){
			fprintf( stderr, "unable to allocate space for xold\n" );
			exit(1);
		}
	}

	/*  space for holding square root of mass */
	if( (gamma_ln != 0.0) && !sqrmass ){
		sqrmass = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( sqrmass == NULL ){
			fprintf( stderr, "unable to allocate space for sqrmass\n" );
			exit(1);
		}
	}

	/*  space for holding the accelerations */
	if( !accel ){
		accel = (REAL_T *) calloc(n, sizeof(REAL_T));
		if( accel == NULL ){
			fprintf( stderr, "unable to allocate space for accel\n" );
			exit(1);
		}
	}

	dtx = dt * 20.455;
	dt5 = 0.5 * dtx;
	rndf = n - 3 * nfrozen;
	if( irattle ) rndf -= prm->Nbonh + prm->Mbona;
	ekin0 = boltz2 * rndf * temp0;	/* target kinetic energy  */
	dttp = dt / tautp;
	zero = 0.0;
	gammai = gamma_ln/20.455;
	c_implic = 1.0/(1.0+gammai*dt5);
	c_explic = 1.0 - gammai*dt5;
	sdfac = sqrt( 4.0*gammai*boltz2*temp0/dtx );

	invmass = 1. / genmass;
	if (prm){
		get_masses(minv);
	} else { /* if no prmtop file is present, see all masses to genmass */
		for (i = 0; i < n; i++) minv[i] = invmass;
	}

	if (zerov) {
		for (i = 0; i < n; i++)
			v[i] = 0.0;
		ekin = 0.0;
	} else if (tempi > 0.0) {
		ekin = 0.0;
		for (i = 0; i < n; i++) {
			if ((prm) && (frozen[i / dim])) {
				v[i] = 0.0;
			} else {
				sd = sqrt(2. * boltz2 * tempi * minv[i]);
				v[i] = gauss(&zero, &sd, &idum);
				ekin += v[i] * v[i] / minv[i];
			}
		}
		ekin *= 0.5;
		temp = ekin / (boltz2 * rndf);

	} else {
		for (ekin = 0., i = 0; i < n; i++)
			ekin += v[i] * v[i] * minv[i];
		ekin *= 0.5;
	}

/*  main loop for the dynamics:  Use velocity Verlet algorithm  */

	nstep = 0;
	epot = mme(x, f, &nstep);

	if( gammai == 0.0 ){  
		for (i = 0; i < n; i++){
			accel[i] = -f[i] * minv[i] * dt5;
		}
	} else {
		for (i = 0; i < n; i++){
			sqrmass[i] = 1.0/sqrt(minv[i]);
			sd = sdfac*sqrmass[i];
			accel[i] = (-f[i] + gauss(&zero, &sd, &idum)) * minv[i] * dt5;
		}
	}


	for (nstep = 1; nstep <= maxstep; nstep++) {

		if (ekin > 0.01)
			tscal = sqrt(1. + dttp * (ekin0 / ekin - 1.));
		else
			tscal = 1.0;
		ekin = 0.0;

		if( irattle )
			for (i = 0; i < n; i++) xold[i] = x[i];

		/*  update v[i] by half-timestep, then x[] by a full timestep:  */
		if( gammai == 0.0 ){
			for (i = 0; i < n; i++) {
				v[i] = (v[i] + accel[i]) * tscal;
				v[i] = v[i] > vlimit ? vlimit : v[i];
				v[i] = v[i] < -vlimit ? -vlimit : v[i];
				x[i] += v[i] * dtx;
			}
		} else {
			for (i = 0; i < n; i++) {
				v[i] = c_explic*v[i] + accel[i];
				v[i] = v[i] > vlimit ? vlimit : v[i];
				v[i] = v[i] < -vlimit ? -vlimit : v[i];
				x[i] += v[i] * dtx;
			}
		}

		if( irattle ) rattle( dtx, x, xold, v, minv );

		/*  get the new energies and forces:  */
		epot = mme(x, f, &nstep);

		/*  update the velocities by the second half-timestep:  */
		if( gammai == 0.0 ){
			for (i = 0; i < n; i++){
				accel[i] = -f[i] * minv[i] * dt5;
				v[i] = (v[i] + accel[i]) * tscal;
			}
		} else {
			for (i = 0; i < n; i++){
				sd = sdfac*sqrmass[i];
				accel[i] = (-f[i] + gauss(&zero, &sd, &idum)) * minv[i] * dt5;
				v[i] = (v[i] + accel[i]) * c_implic;
			}
		}

		if( irattle ) rattle2( dtx, x, v, minv );

		/*  get the kinetic and total energies; update the time:  */
		for (i = 0; i < n; i++) ekin += v[i]*v[i]/minv[i];
		ekin *= 0.5;
		etot = ekin + epot;
		temp = ekin / (boltz2 * rndf);
		t += dt;

		if (nstep % ntpr_md == 0 || nstep == 1) {
			printf("md:       %5d %10.3f %10.2f %10.2f %10.2f %10.2f\n",
				   nstep, t, ekin, epot, ekin + epot, temp);
			fflush(stdout);
		}
		if (ntwx > 0 && nstep % ntwx == 0 && binposfp != NULL)
			writebinposfrm(n / 3, x, binposfp);
	}
	return (0);
}

int	mme_timer( void )
{
	printf("\nTiming summary:\n");
	printf("   bonds       %8.1f\n", tbond);
	printf("   angles      %8.1f\n", tangl);
	printf("   torsions    %8.1f\n", tphi);
	printf("   pairlist    %8.1f\n", tpair);
	printf("   nonbonds    %8.1f\n", tnonb);
	printf("   constraints %8.1f\n", tcons);
	printf("   Total       %8.1f\n\n",
		   tbond + tangl + tphi + tpair + tnonb + tcons);
	return (0);
}

/*   #include "shiftnbond.c"   */
#include "rattle.c"
#include "lex.mm_options.c"
